///////////////////////////////////////////////////////////////////////////
// Phoenix Emulator v2.1 by D. Finck                                     //
///////////////////////////////////////////////////////////////////////////

///////////////////////Header Files///////////////////////
#include <stdio.h>                              //for most anything
#include <time.h>                               //for framerate timer
#include "z80.h"                                //for the CPU module
#include "allegro.h"                            //for the Allegro library
DECLARE_GFX_DRIVER_LIST( GFX_DRIVER_MODEX )     //these help keep the .exe
DECLARE_COLOR_DEPTH_LIST( COLOR_DEPTH_8 )       //file size to a minimum
                                                //(read about it in the Allegro docs)
///////////////////////Function Prototypes////////////////
void StartMessage(void);
void AllocateMem(void);
void ReadROMs(void);                                    
void SetupPalletes(void);
void GameInput(void);
void EmuInput(void);
void DrawScreen(void);                           
void DrawCharA(int);
void DrawCharB(int);
void LoadRom(char*, char*, word, word);

///////////////////////Global Variables///////////////////
//(global variables are cool...)
int Z80_Iperiod=720000/60;      //cycles per vblank (used by Z80 module)
int Z80_IRQ=0;                  //(used by Z80 module)
byte Dips = 0x43;               //settings of hardware dips
byte *RAM;			//declare global "RAM"
byte *CharA;                    //character data A
byte *CharB;                    //character data B
byte *VRAM0;                    //video RAM, bank 0
byte *VRAM1;                    //video RAM, bank 1
int VRAM = 0;                   //current VRAM bank
BITMAP *bm1;                    //buffer bitmap 1-holds foreground pic
BITMAP *bm2;                    //buffer bitmap 2-holds background pic
BITMAP *bm3;                    //buffer bitmap 3-holds combined pic
BITMAP *bmsc;                   //sub-bitmap for screen capture
PALLETE pal1;                   //pallete 1
PALLETE pal2;                   //pallete 2
PALLETE palsc;                  //pallete for screen capture
uclock_t clk1,clk2;             //timers for monitoring & limiting speed
int DirtyCharA[832];            //buffer for changes in video A
int DirtyCharB[832];            //buffer for changes in video B
int fps = 0;                    //for monitoring speed
int throttle = 1;               //for limiting speed
int fskipmax = 3;               //maximum frameskip
int frameskip = 0;              //default frameskip setting
int skip=0;                     //frameskip counter
char msg[80];                   //for printing text to screen
FILE *fp;                       //handy file pointer
int tst=0;                      //counter for hi-score loading
int Game=1;                     //1=Phoenix (default), 2=Plieads
char *HiFile;                   //filename for hi-score

////////////////////////////////////////////////////////////
// StartMessage prints a message to the user upon startup //
////////////////////////////////////////////////////////////
void StartMessage() {
  printf("\nKeys:\n");
  printf(" LEFT      = left\n");
  printf(" RIGHT     = right\n");
  printf(" LEFT CTRL = shield\n");
  printf(" LEFT ALT  = fire\n");
  printf(" P         = pause\n");
  printf(" F3        = reset\n");
  printf(" F8        = frameskip\n");
  printf(" F10       = speed throttle\n");
  printf(" F11       = show framerate\n");
  printf(" F12       = save screen shot\n\n");
  printf("Center Joystick and press a KEY\n");//prompt user...
  getch();                                    //...and wait for a key
//  joy_type=JOY_TYPE_4BUTTON;                  //should work for most joys
  initialise_joystick();                      //Allegro function
  return;
}

//////////////////////////////////////////////////////////
//                      M A I N                         //
//////////////////////////////////////////////////////////
int main(int argc, char **argv) {
  if (argc > 1) Game=2;                       //any argument gives Plieads
  clrscr();                                   //clear the screen
  AllocateMem();                              //allocate memory   
  SetupPalletes();                            //self explanitory
  ReadROMs();                                 //get the ROMs, we'll need 'em
  Z80_Reset();                                //initalize CPU
  StartMessage();                             //info to user
  allegro_init();                             //initalize Allegro
  install_keyboard();                         //Allegro's key handler
  set_gfx_mode(GFX_MODEX,360,270,0,0);        //set graphics mode
  set_pallete(pal1);                          //set the pallete
  clear(screen);                              //and clear the screen
  delay(500);                                 //wait half a sec
  clk1=uclock();                              //initalize FPS timer
  Z80();                                      //let the emulation begin! (:
  printf("Exiting Emulator...\n\n");          //now for the graceful exit
  free(RAM);free(CharA);free(CharB);          //un-allocate memory
  return 0;                                   //let the emulation end. ):
}

/////////////////////////////////////////////////////////////////////
// Z80_WRMEM handles Writing to "Memory", as required by the "CPU" //
/////////////////////////////////////////////////////////////////////
void Z80_WRMEM(dword A, byte V) {
  if (A>=0x4000 && A<0x4400) {                //
    if (A>=0x4000 && A<0x4340)                //writing to video A?
      DirtyCharA[A-0x4000]=1;                 //keep track of changes
    if (VRAM) VRAM1[A-0x4000]=V;              //handle video RAM banking
    else VRAM0[A-0x4000]=V;                   //on the memory-write end
  }
  else if (A>=0x4800 && A<0x4C00) {           //
    if (A>=0x4800 && A<0x4B40)                //writing to video B?
      DirtyCharB[A-0x4800]=1;                 //track changes
    if (VRAM) VRAM1[A-0x4000]=V;              //handle video RAM banking
    else VRAM0[A-0x4000]=V;                   //on the memory-write end
  }
  else if (A>=0x5800 && A<0x5C00) RAM[0x5800]=V; //mirror scroll register
  else if (A>=0x5000 && A<0x5400 && (RAM[0x5000]&1)!=(V&1)) { //if changing this bit...
    int count;                                //(will need counter var)
    if (V&1) VRAM=1;                          //set VRAM bank
    else VRAM=0;                              //appropriately
    for (count=0;count<832;count++) {         //mark every byte of
      DirtyCharA[count]=1;                    //screen RAM "dirty"
      DirtyCharB[count]=1;                    //to force full re-draw
    }
    RAM[0x5000]=(RAM[0x5000]&~1)+(V&1);       //write to base addr
  }
  if (A>=0x5000 && A<0x5400 && (RAM[0x5000]&2)!=(V&2)) { //if changing this bit...
    if (V&2)                                  //raised bit 1?
      set_pallete(pal2);                      //select alternate pallete
    else                                      //must've unset bit 1...
      set_pallete(pal1);                      //select primary pallete
    RAM[0x5000]=(RAM[0x5000]&~2)+(V&2);       //store byte at "base" addr
  }
  if (A<0x4000||A>=0x8000) return;            //don't write on ROMs...
  RAM[A]=V;                                   //OK, put byte in memory
  return;               
}

//////////////////////////////////////////////////////////////////////
// Z80_RDMEM handles Fetching Data Stored in "Memory" for the "CPU" //
//////////////////////////////////////////////////////////////////////
unsigned Z80_RDMEM(dword A) {
  if (A<0x4000) return RAM[A];                //should be most common case
  if (A>=0x4000 && A<0x4400) {           //this deal here handles
    if (VRAM) return(VRAM1[A-0x4000]);        //banking the video RAM
    else return(VRAM0[A-0x4000]);             //from the mem-read end
  }
  if (A>=0x4800 && A<0x4C00) {           //this deal here handles
    if (VRAM) return(VRAM1[A-0x4000]);        //banking the video RAM
    else return(VRAM0[A-0x4000]);             //from the mem-read end
  }
  if (A>=0x7800 && A<0x7C00) {           //reading dipswitches?
    if(RAM[0x7800]&0x80) {                    //if sync bit of switch high
      RAM[0x7800]&=~0x80;                     //clear the bit
      return (RAM[0x7800]|0x80);              //return value with bit 7 high
    }
    return RAM[0x7800];                  //return value from memory
  }
  if (A>=0x7000 && A<0x7400) {           //if reading input area...
    GameInput();                              //..get some input...
    return RAM[0x7000];                       //...and return this byte
  }                                           //(this achieves "mirroring")
  return RAM[A];                              //return the value from RAM
}

///////////////////////////////////////////
// DrawScreen handles drawing the screen //
///////////////////////////////////////////
void DrawScreen() {
  int pos,temp;                               //declare the vars needed...
  int scroll = RAM[0x5800];                   //scroll row position
  for (pos=0;pos<832;pos++) {                 //for each byte in video
    if (DirtyCharA[pos]) {                    //if is is marked...
      DirtyCharA[pos]=0;                      //...unmark it...
      DrawCharA(pos);                         //...and draw to bitmap
    }
    if (DirtyCharB[pos]) {                    //also do the same for
      DirtyCharB[pos]=0;                      //the background plane
      DrawCharB(pos);                         //while we're at it
    }
  }
  temp=256-scroll;                            //split line
  set_clip(bm3,0,8,208,256);                  //dont show top 8 raster lines
  blit(bm2,bm3,0,scroll,0,0,208,temp);        //blit scrolling bg...
  blit(bm2,bm3,0,0,0,temp,208,scroll);        //...in two parts
  set_clip(bm3,0,0,208,256);                  //allow whole bitmap now
  masked_blit(bm1,bm3,0,0,0,0,208,256);       //blit foregrnd over bkgrnd
  blit(bm3,screen,0,0,76,0,208,256);          //finally to the screen
  if (throttle) {
    while (uclock()-clk1 < (frameskip+1)*UCLOCKS_PER_SEC/61) { } //throttle speed
  }
  if (fps) {
    sprintf(msg,"FPS=%-3d",(frameskip+1)*UCLOCKS_PER_SEC/(uclock()-clk1));//display framerate
    textout(screen,font,msg,SCREEN_W/2+45,SCREEN_H-10,2);
  }
  clk1=uclock();                            //reset the timer variable
//  sprintf(msg,"%d",Game);        //display on screen
//  textout_centre(screen,font,msg,SCREEN_W/2-120,SCREEN_H-10,2);//for debugging
  return;
}

//////////////////////////////////////////////////////
// DrawCharA draws one character to buffer bitmap 1 //
//////////////////////////////////////////////////////
void DrawCharA(int pos) {
  int x = 25-(pos/32);                          //calculate x and
  int y = pos%32;                               //y coordinates of pos
  unsigned int cx,cy;                           //local pixel coords
  unsigned char chr,b1,b2,col,col2;             //other vars
  if (VRAM) chr=VRAM1[(25-x)*32+y];             //char code from proper
  else chr=VRAM0[(25-x)*32+y];                  //bank of VRAM
  col=4*(chr>>5);                               //calc "base" color index
  for (cx=0;cx<8;cx++) {                        //loop through columns
    b1=CharA[chr*8+cx];                         //byte from char ROMs
    b2=CharA[chr*8+cx+0x0800];                  //byte from char ROMs
    for (cy=0;cy<8;cy++) {                      //loop through each bit
        col2=col+2*((b1>>cy)&1);                //adjust color index
        col2+=(b2>>cy)&1;                       //accordingly
        if (col2%4==0) col2=0;                  //for "transparancy" of fg
        bm1->line[(y*8)+cy][(x*8)+(7-cx)]=col2; //direct memory access
    }                                    
  }         
}

//////////////////////////////////////////////////////
// DrawCharB draws one character to buffer bitmap 2 //
//////////////////////////////////////////////////////
void DrawCharB(int pos) {
  int x = 25-(pos/32);                          //calculate x and
  int y = pos%32;                               //y char coordinates
  unsigned int cx,cy;                           //local pixel coords
  unsigned int chr,b1,b2,col,col2;              //other variables
  if (VRAM) chr=VRAM1[0x0800+(25-x)*32+y];      //char code from proper
  else chr=VRAM0[0x0800+(25-x)*32+y];           //bank of VRAM
  col=4*(chr>>5)+32;                            //calc "base" color
  for (cx=0;cx<8;cx++) {                        //loop through columns
    b1=CharB[chr*8+cx];                         //bytes from char ROMs
    b2=CharB[chr*8+cx+0x0800];                  //other bitplane
    for (cy=0;cy<8;cy++) {                      //loop through each bit
        col2=col+2*((b1>>cy)&1);                //adjust color per
        col2+=(b2>>cy)&1;                       //bits from char ROMs
        bm2->line[(y*8)+cy][(x*8)+(7-cx)]=col2; //draw pixel on bitmap
    }                                           
  }
}

//////////////////////////////////////////////////
// GameInput retrieves and acts upon all inputs //
//////////////////////////////////////////////////
void GameInput() {
  poll_joystick();                            //update joystick variables
  RAM[0x7000]=0xFF;                           //clear input (all bits high)
  if(key[KEY_3]) RAM[0x7000]&=~0x01;          //insert coin      (bit 0 low)
  if(key[KEY_1]) RAM[0x7000]&=~0x02;          //press 1PL button (bit 1 low)
  if(key[KEY_2]) RAM[0x7000]&=~0x04;          //press 2PL button (bit 2 low)
  if(key[KEY_ALT]     || joy_b1)
                 RAM[0x7000]&=~0x10;          //press FIRE button(bit 4 low)
  if(key[KEY_RIGHT]   || joy_right || joy_x> 50)
                 RAM[0x7000]&=~0x20;          //joystick right   (bit 5 low)
  if(key[KEY_LEFT]    || joy_left  || joy_x<(-50))
                 RAM[0x7000]&=~0x40;          //joystick left    (bit 6 low)
  if(key[KEY_CONTROL] || joy_b2)
                 RAM[0x7000]&=~0x80;          //shield button    (bit 7 low)
}

///////////////////////////////////////////
// EmuInput
/////////////////////////////////////////
void EmuInput() {
  if(key[KEY_ESC]) {                          //shutdown on ESC
    byte count;                               
    Z80_Running=0;                            
    fp=fopen(HiFile,"w");              //but save hi-score first
    for(count=0;count<4;count++)              //four bytes from this location
      putc(VRAM0[0x388+count],fp);            //seem to sort-of do it
    fclose(fp);                               
  }
  if(key[KEY_F8]) {                           //if F8 is pressed...
    frameskip++;                              //increment frameskip
    if(frameskip>fskipmax) frameskip=0;       //wraparound at max
    sprintf(msg,"FSKIP=%d",frameskip);        //display on screen
    textout_centre(screen,font,msg,SCREEN_W/2-70,SCREEN_H-10,2);
    clear_keybuf();                           //clear keyboard buffer
  }
  if(key[KEY_P]) {                            //if P is pressed...
    textout_centre(screen, font, "PAUSED",SCREEN_W/2,SCREEN_H-10,2);
    clear_keybuf();
    readkey();                                //wait for another keypress
    textout_centre(screen, font, "      ",SCREEN_W/2,SCREEN_H-10,0);
    clear_keybuf();
  }
  if(key[KEY_F11]) {                          //if F11 is pressed
    fps=1-fps;                                //toggle fps
    if(!fps)                                  //if it was cut off...
      textout(screen, font, "        ",SCREEN_W/2+45,SCREEN_H-10,0);//clear the words
    clear_keybuf();                           //clear the keyboard buffer
  }
  if(key[KEY_F3]) {                           //if F3 is pressed
    Z80_Reset();                              //reset the CPU
    clear_keybuf();
  }
  if(key[KEY_F10]) {                          //if F10 is pressed
    throttle=1-throttle;                      //toggle throttling
    clear_keybuf();                           //clear the keyboard buffer
  }
  if(key[KEY_F12]) {                          //captures a screenshot
    get_pallete(palsc);                       //and saves it to a file
    bmsc=create_sub_bitmap(screen,76,0,208,256);//sub-bitmap insures proper
    save_bitmap("screen.bmp", bmsc, palsc);   //size of graphic
    destroy_bitmap(bmsc);                     //clean up
  }
  return;
}

/////////////////////////////////////////////////////////////
// AllocateMem allocates memory ROM data and video buffers //
/////////////////////////////////////////////////////////////
void AllocateMem() {
  word count;                                 //we'll need a variable
  printf("Allocating Memory\n");              //be polite
  RAM   = malloc(0x8000);                     //allocate memory for "RAM"
  CharA = malloc(0x1000);                     //and for gfx characters
  CharB = malloc(0x1000);                     //and for more gfx data
  VRAM0 = malloc(0x1000);                     //and for video bank0
  VRAM1 = malloc(0x1000);                     //and for video bank1
  if (!RAM||!CharA||!CharB||!VRAM0||!VRAM1) { //if those didn't work
    printf("Error allocating memory\n");      //mention the problem
    exit(1);                                  //and get outta here
  }
  for (count=0x4000;count<0x8000;count++)     //
    RAM[count]=0;                             //fill memory with 0's
  for (count=0x0000;count<0x1000;count++) {   //
    VRAM0[count]=0;                           //also fill both banks
    VRAM1[count]=0;                           //of VRAM with 0's
  }    
  bm1=create_bitmap(208,256);                 //create a bitmaps
  bm2=create_bitmap(208,256);                 //for 2 gfx planes and
  bm3=create_bitmap(208,256);                 //final image
  clear(bm1); clear(bm2); clear(bm3);         //clear them all
  return;
}

/////////////////////////////////////////////
// ReadROMs reads the game ROMs from files //
/////////////////////////////////////////////
void ReadROMs() {
  if (Game==1) {
    printf("Reading Phoenix ROMs...\n");
    LoadRom("phoenix/ic45", RAM, 0x0000, 0x800);
    LoadRom("phoenix/ic46", RAM, 0x0800, 0x800);  
    LoadRom("phoenix/ic47", RAM, 0x1000, 0x800);
    LoadRom("phoenix/ic48", RAM, 0x1800, 0x800);  
    LoadRom("phoenix/ic49", RAM, 0x2000, 0x800);
    LoadRom("phoenix/ic50", RAM, 0x2800, 0x800);  
    LoadRom("phoenix/ic51", RAM, 0x3000, 0x800);
    LoadRom("phoenix/ic52", RAM, 0x3800, 0x800);
    LoadRom("phoenix/ic39", CharA, 0x0000, 0x800);
    LoadRom("phoenix/ic40", CharA, 0x0800, 0x800);  
    LoadRom("phoenix/ic23", CharB, 0x0000, 0x800);
    LoadRom("phoenix/ic24", CharB, 0x0800, 0x800);
    RAM[0x7800]=Dips;                             //set dipswitches
    HiFile = "phoenix/phoenix.hi";
  }
  if (Game==2) {
  printf("Reading Plieads ROMs...\n");
    LoadRom("plieads/ic45", RAM, 0x0000, 0x800);
    LoadRom("plieads/ic46", RAM, 0x0800, 0x800);  
    LoadRom("plieads/ic47", RAM, 0x1000, 0x800);
    LoadRom("plieads/ic48", RAM, 0x1800, 0x800);  
    LoadRom("plieads/ic49", RAM, 0x2000, 0x800);
    LoadRom("plieads/ic50", RAM, 0x2800, 0x800);  
    LoadRom("plieads/ic51", RAM, 0x3000, 0x800);
    LoadRom("plieads/ic52", RAM, 0x3800, 0x800);
    LoadRom("plieads/ic39", CharA, 0x0000, 0x800);
    LoadRom("plieads/ic40", CharA, 0x0800, 0x800);  
    LoadRom("plieads/ic23", CharB, 0x0000, 0x800);
    LoadRom("plieads/ic24", CharB, 0x0800, 0x800);
    RAM[0x7800]=Dips;                             //set dipswitches
    HiFile = "plieads/plieads.hi";
  }
  return;
}

//////////////////////////////////////////////////////////
// LoadRom loads one ROM into the given memory location //
//////////////////////////////////////////////////////////
void LoadRom(char *name, char *loc, word pos, word len)
{
  fp=fopen(name, "rb");
  if(!fp) {
    printf("Error loading ROM: %s\n",name);
    exit(1);
  }
  fread(&loc[pos], len, 1, fp);
  fclose(fp);
  return;
}

void SetupPalletes() {
//Pallete 1
  pal1[0].r=0x00>>2;pal1[0].g=0x00>>2; pal1[0].b=0x00>>2; //black
  pal1[1].r=0x00>>2;pal1[1].g=0x00>>2; pal1[1].b=0x00>>2; //black
  pal1[2].r=0x00>>2;pal1[2].g=0xFF>>2; pal1[2].b=0xDB>>2; //cyan
  pal1[3].r=0x00>>2;pal1[3].g=0xFF>>2; pal1[3].b=0xDB>>2; //cyan
  pal1[4].r=0x00>>2;pal1[4].g=0x00>>2; pal1[4].b=0x00>>2; //black
  pal1[5].r=0xFF>>2;pal1[5].g=0xFF>>2; pal1[5].b=0x00>>2; //yellow
  pal1[6].r=0xFF>>2;pal1[6].g=0x00>>2; pal1[6].b=0x00>>2; //red
  pal1[7].r=0xDB>>2;pal1[7].g=0xDB>>2; pal1[7].b=0xDB>>2; //white
  pal1[8].r=0x00>>2;pal1[8].g=0x00>>2; pal1[8].b=0x00>>2; //black
  pal1[9].r=0xFF>>2;pal1[9].g=0xFF>>2; pal1[9].b=0x00>>2; //yellow
  pal1[10].r=0xFF>>2;pal1[10].g=0x00>>2; pal1[10].b=0x00>>2; //red
  pal1[11].r=0xDB>>2;pal1[11].g=0xDB>>2; pal1[11].b=0xDB>>2; //white
  pal1[12].r=0x00>>2;pal1[12].g=0x00>>2; pal1[12].b=0x00>>2; //black
  pal1[13].r=0xFF>>2;pal1[13].g=0xB6>>2; pal1[13].b=0xDB>>2; //pink
  pal1[14].r=0xFF>>2;pal1[14].g=0x00>>2; pal1[14].b=0xFF>>2; //purple
  pal1[15].r=0xFF>>2;pal1[15].g=0xFF>>2; pal1[15].b=0x00>>2; //yellow
  pal1[16].r=0x00>>2;pal1[16].g=0x00>>2; pal1[16].b=0x00>>2; //black
  pal1[17].r=0xFF>>2;pal1[17].g=0xB6>>2; pal1[17].b=0xDB>>2; //pink
  pal1[18].r=0xFF>>2;pal1[18].g=0x00>>2; pal1[18].b=0xFF>>2; //purple
  pal1[19].r=0xFF>>2;pal1[19].g=0xFF>>2; pal1[19].b=0x00>>2; //yellow
  pal1[20].r=0x00>>2;pal1[20].g=0x00>>2; pal1[20].b=0x00>>2; //black
  pal1[21].r=0xFF>>2;pal1[21].g=0xB6>>2; pal1[21].b=0xDB>>2; //pink
  pal1[22].r=0xFF>>2;pal1[22].g=0x00>>2; pal1[22].b=0xFF>>2; //purple
  pal1[23].r=0xFF>>2;pal1[23].g=0xFF>>2; pal1[23].b=0x00>>2; //yellow
  pal1[24].r=0x00>>2;pal1[24].g=0x00>>2; pal1[24].b=0x00>>2; //black
  pal1[25].r=0xDB>>2;pal1[25].g=0xDB>>2; pal1[25].b=0xDB>>2; //white
  pal1[26].r=0xFF>>2;pal1[26].g=0x00>>2; pal1[26].b=0xFF>>2; //purple
  pal1[27].r=0xFF>>2;pal1[27].g=0xFF>>2; pal1[27].b=0x00>>2; //yellow
  pal1[28].r=0x00>>2;pal1[28].g=0x00>>2; pal1[28].b=0x00>>2; //black
  pal1[29].r=0xFF>>2;pal1[29].g=0x00>>2; pal1[29].b=0xFF>>2; //purple
  pal1[30].r=0x00>>2;pal1[30].g=0xFF>>2; pal1[30].b=0x00>>2; //green
  pal1[31].r=0xDB>>2;pal1[31].g=0xDB>>2; pal1[31].b=0xDB>>2; //white
  pal1[32].r=0x00>>2;pal1[32].g=0x00>>2; pal1[32].b=0x00>>2; //black
  pal1[33].r=0xFF>>2;pal1[33].g=0x00>>2; pal1[33].b=0x00>>2; //red
  pal1[34].r=0x24>>2;pal1[34].g=0x24>>2; pal1[34].b=0xDB>>2; //blue
  pal1[35].r=0xDB>>2;pal1[35].g=0xDB>>2; pal1[35].b=0xDB>>2; //white
  pal1[36].r=0x00>>2;pal1[36].g=0x00>>2; pal1[36].b=0x00>>2; //black
  pal1[37].r=0xFF>>2;pal1[37].g=0x00>>2; pal1[37].b=0xFF>>2; //purple
  pal1[38].r=0x95>>2;pal1[38].g=0x95>>2; pal1[38].b=0xFF>>2; //blueish
  pal1[39].r=0xFF>>2;pal1[39].g=0xB6>>2; pal1[39].b=0x00>>2; //dark orange
  pal1[40].r=0x00>>2;pal1[40].g=0x00>>2; pal1[40].b=0x00>>2; //black
  pal1[41].r=0xB6>>2;pal1[41].g=0x24>>2; pal1[41].b=0xFF>>2; //dark purple
  pal1[42].r=0x00>>2;pal1[42].g=0xFF>>2; pal1[42].b=0x00>>2; //green
  pal1[43].r=0xFF>>2;pal1[43].g=0xB6>>2; pal1[43].b=0x00>>2; //dark orange
  pal1[44].r=0x00>>2;pal1[44].g=0x00>>2; pal1[44].b=0x00>>2; //black
  pal1[45].r=0x95>>2;pal1[45].g=0x95>>2; pal1[45].b=0xFF>>2; //blueish
  pal1[46].r=0xB6>>2;pal1[46].g=0x24>>2; pal1[46].b=0xFF>>2; //dark purple
  pal1[47].r=0xFF>>2;pal1[47].g=0x24>>2; pal1[47].b=0xB6>>2; //light purple
  pal1[48].r=0x00>>2;pal1[48].g=0x00>>2; pal1[48].b=0x00>>2; //black
  pal1[49].r=0xFF>>2;pal1[49].g=0x00>>2; pal1[49].b=0xFF>>2; //purple
  pal1[50].r=0x95>>2;pal1[50].g=0x95>>2; pal1[50].b=0xFF>>2; //blueish
  pal1[51].r=0x00>>2;pal1[51].g=0xFF>>2; pal1[51].b=0x00>>2; //green
  pal1[52].r=0x00>>2;pal1[52].g=0x00>>2; pal1[52].b=0x00>>2; //black
  pal1[53].r=0xFF>>2;pal1[53].g=0x00>>2; pal1[53].b=0xFF>>2; //purple
  pal1[54].r=0x95>>2;pal1[54].g=0x95>>2; pal1[54].b=0xFF>>2; //blueish
  pal1[55].r=0x00>>2;pal1[55].g=0xFF>>2; pal1[55].b=0x00>>2; //green
  pal1[56].r=0x00>>2;pal1[56].g=0x00>>2; pal1[56].b=0x00>>2; //black
  pal1[57].r=0xFF>>2;pal1[57].g=0x00>>2; pal1[57].b=0xFF>>2; //purple
  pal1[58].r=0x95>>2;pal1[58].g=0x95>>2; pal1[58].b=0xFF>>2; //blueish
  pal1[59].r=0x00>>2;pal1[59].g=0xFF>>2; pal1[59].b=0x00>>2; //green
  pal1[60].r=0x00>>2;pal1[60].g=0x00>>2; pal1[60].b=0x00>>2; //black
  pal1[61].r=0xFF>>2;pal1[61].g=0x00>>2; pal1[61].b=0xFF>>2; //purple
  pal1[62].r=0x95>>2;pal1[62].g=0x95>>2; pal1[62].b=0xFF>>2; //blueish
  pal1[63].r=0x00>>2;pal1[63].g=0xFF>>2; pal1[63].b=0x00>>2; //green
//Pallete 2
  pal2[0].r=0x00>>2;pal2[0].g=0x00>>2; pal2[0].b=0x00>>2; //black
  pal2[1].r=0x24>>2;pal2[1].g=0x24>>2; pal2[1].b=0xDB>>2; //black
  pal2[2].r=0x00>>2;pal2[2].g=0xFF>>2; pal2[2].b=0xDB>>2; //cyan
  pal2[3].r=0x00>>2;pal2[3].g=0xFF>>2; pal2[3].b=0xDB>>2;//cyan
  pal2[4].r=0x00>>2;pal2[4].g=0x00>>2; pal2[4].b=0x00>>2; //black
  pal2[5].r=0xFF>>2;pal2[5].g=0xFF>>2; pal2[5].b=0x00>>2; //yellow
  pal2[6].r=0xFF>>2;pal2[6].g=0x00>>2; pal2[6].b=0x00>>2; //red
  pal2[7].r=0xDB>>2;pal2[7].g=0xDB>>2; pal2[7].b=0xDB>>2; //white
  pal2[8].r=0x00>>2;pal2[8].g=0x00>>2; pal2[8].b=0x00>>2; //black
  pal2[9].r=0xFF>>2;pal2[9].g=0xFF>>2; pal2[9].b=0x00>>2; //yellow
  pal2[10].r=0xFF>>2;pal2[10].g=0x00>>2; pal2[10].b=0x00>>2; //red
  pal2[11].r=0xDB>>2;pal2[11].g=0xDB>>2; pal2[11].b=0xDB>>2; //white
  pal2[12].r=0x00>>2;pal2[12].g=0x00>>2; pal2[12].b=0x00>>2; //black
  pal2[13].r=0xFF>>2;pal2[13].g=0xFF>>2; pal2[13].b=0x00>>2; //yellow
  pal2[14].r=0x00>>2;pal2[14].g=0xFF>>2; pal2[14].b=0x00>>2; //green
  pal2[15].r=0xFF>>2;pal2[15].g=0x00>>2; pal2[15].b=0xFF>>2; //purple
  pal2[16].r=0x00>>2;pal2[16].g=0x00>>2; pal2[16].b=0x00>>2; //black
  pal2[17].r=0xFF>>2;pal2[17].g=0xFF>>2; pal2[17].b=0x00>>2; //yellow
  pal2[18].r=0x00>>2;pal2[18].g=0xFF>>2; pal2[18].b=0x00>>2; //green
  pal2[19].r=0xFF>>2;pal2[19].g=0x00>>2; pal2[19].b=0xFF>>2; //purple
  pal2[20].r=0x00>>2;pal2[20].g=0x00>>2; pal2[20].b=0x00>>2; //black
  pal2[21].r=0xFF>>2;pal2[21].g=0xFF>>2; pal2[21].b=0x00>>2; //yellow
  pal2[22].r=0x00>>2;pal2[22].g=0xFF>>2; pal2[22].b=0x00>>2; //green
  pal2[23].r=0xFF>>2;pal2[23].g=0x00>>2; pal2[23].b=0xFF>>2; //purple
  pal2[24].r=0x00>>2;pal2[24].g=0x00>>2; pal2[24].b=0x00>>2; //black
  pal2[25].r=0xDB>>2;pal2[25].g=0xDB>>2; pal2[25].b=0xDB>>2; //white
  pal2[26].r=0xFF>>2;pal2[26].g=0x00>>2; pal2[26].b=0x00>>2; //red
  pal2[27].r=0xFF>>2;pal2[27].g=0x00>>2; pal2[27].b=0xFF>>2; //purple
  pal2[28].r=0x00>>2;pal2[28].g=0x00>>2; pal2[28].b=0x00>>2; //black
  pal2[29].r=0xFF>>2;pal2[29].g=0x00>>2; pal2[29].b=0xFF>>2; //purple
  pal2[30].r=0x00>>2;pal2[30].g=0xFF>>2; pal2[30].b=0x00>>2; //green
  pal2[31].r=0xDB>>2;pal2[31].g=0xDB>>2; pal2[31].b=0xDB>>2; //white
  pal2[32].r=0x00>>2;pal2[32].g=0x00>>2; pal2[32].b=0x00>>2; //black
  pal2[33].r=0xFF>>2;pal2[33].g=0x00>>2; pal2[33].b=0x00>>2; //red
  pal2[34].r=0x24>>2;pal2[34].g=0x24>>2; pal2[34].b=0xDB>>2; //blue
  pal2[35].r=0xDB>>2;pal2[35].g=0xDB>>2; pal2[35].b=0xDB>>2; //white
  pal2[36].r=0x00>>2;pal2[36].g=0x00>>2; pal2[36].b=0x00>>2; //black
  pal2[37].r=0xFF>>2;pal2[37].g=0x00>>2; pal2[37].b=0xFF>>2; //purple
  pal2[38].r=0x95>>2;pal2[38].g=0x95>>2; pal2[38].b=0xFF>>2; //blueish
  pal2[39].r=0xFF>>2;pal2[39].g=0xB6>>2; pal2[39].b=0x00>>2; //dark orange
  pal2[40].r=0x00>>2;pal2[40].g=0x00>>2; pal2[40].b=0x00>>2; //black
  pal2[41].r=0xB6>>2;pal2[41].g=0x24>>2; pal2[41].b=0xFF>>2; //dark purple
  pal2[42].r=0x00>>2;pal2[42].g=0xFF>>2; pal2[42].b=0x00>>2; //green
  pal2[43].r=0xFF>>2;pal2[43].g=0xB6>>2; pal2[43].b=0x00>>2; //dark orange
  pal2[44].r=0x00>>2;pal2[44].g=0x00>>2; pal2[44].b=0x00>>2; //black
  pal2[45].r=0x95>>2;pal2[45].g=0x95>>2; pal2[45].b=0xFF>>2; //blueish
  pal2[46].r=0xB6>>2;pal2[46].g=0x24>>2; pal2[46].b=0xFF>>2; //dark purple
  pal2[47].r=0xFF>>2;pal2[47].g=0x24>>2; pal2[47].b=0xB6>>2; //light purple
  pal2[48].r=0x00>>2;pal2[48].g=0x00>>2; pal2[48].b=0x00>>2; //black
  pal2[49].r=0x95>>2;pal2[49].g=0x95>>2; pal2[49].b=0xFF>>2; //blueish
  pal2[50].r=0xFF>>2;pal2[50].g=0x24>>2; pal2[50].b=0xB6>>2; //light purple
  pal2[51].r=0x00>>2;pal2[51].g=0xFF>>2; pal2[51].b=0x00>>2; //green
  pal2[52].r=0x00>>2;pal2[52].g=0x00>>2; pal2[52].b=0x00>>2; //black
  pal2[53].r=0x95>>2;pal2[53].g=0x95>>2; pal2[53].b=0xFF>>2; //blueish
  pal2[54].r=0xFF>>2;pal2[54].g=0x24>>2; pal2[54].b=0xB6>>2; //light purple
  pal2[55].r=0x00>>2;pal2[55].g=0xFF>>2; pal2[55].b=0x00>>2; //green
  pal2[56].r=0x00>>2;pal2[56].g=0x00>>2; pal2[56].b=0x00>>2; //black
  pal2[57].r=0x95>>2;pal2[57].g=0x95>>2; pal2[57].b=0xFF>>2; //blueish
  pal2[58].r=0xFF>>2;pal2[58].g=0x24>>2; pal2[58].b=0xB6>>2; //light purple
  pal2[59].r=0x00>>2;pal2[59].g=0xFF>>2; pal2[59].b=0x00>>2; //green
  pal2[60].r=0x00>>2;pal2[60].g=0x00>>2; pal2[60].b=0x00>>2; //black
  pal2[61].r=0x95>>2;pal2[61].g=0x95>>2; pal2[61].b=0xFF>>2; //blueish
  pal2[62].r=0xFF>>2;pal2[62].g=0x24>>2; pal2[62].b=0xB6>>2; //light purple
  pal2[63].r=0x00>>2;pal2[63].g=0xFF>>2; pal2[63].b=0x00>>2; //green
  return;
}

///////////////////////////////////////////////////////////
// Z80_Interrupt handles vblanks, disguised at interupts //
///////////////////////////////////////////////////////////
int Z80_Interrupt() {
  if(!skip--) {                               //skip frames if needed
    skip = frameskip;                         //reset skip counter
    DrawScreen();                             //we'll want to see something
    if (tst<2) {                              //for first 2 vblanks...
      tst++;                                  //increment counter
      fp=fopen(HiFile,"r");            //open hiscore file
      if(fp) {                                //if it exists
        byte count;                           //
        for (count=0;count<4;count++)         //load the data 
          VRAM0[0x384+count]=getc(fp);        //into the proper? place
      }
      fclose(fp);                             //close the file
    }
  }
  EmuInput();
  RAM[0x7800]|=0x80;                          //raise this bit
  return(Z80_IGNORE_INT);                     //and ignore the "interrupt"
}

///////////////////////////////////////////////////////////////////////////////
//The following "do-nothing" functions must exist to satisfy the Z80 emulator//
///////////////////////////////////////////////////////////////////////////////
byte Z80_In(byte Port) {return(0);}
void Z80_Out(byte Port, byte value) {}
void Z80_Patch(Z80_Regs *R) {}
void Z80_Reti() {}
void Z80_Retn() {}
